import { promises, existsSync, statSync } from 'fs';
import { defu } from 'defu';
import { useLogger, defineNuxtModule, createResolver, addPlugin, addTemplate, addImportsDir, extendViteConfig, resolveFiles } from '@nuxt/kit';
import { generate as generate$1 } from '@graphql-codegen/cli';
import { parse } from 'graphql';
import { upperFirst } from 'scule';
import { genExport } from 'knitwork';

const name = "nuxt-graphql-client";
const version = "0.1.25-edge.aecti";

function prepareConfig(options) {
  const prepareSchema = (v) => {
    if (v.schema) {
      v.schema = options.resolver.resolve(v.schema);
      return [v.schema];
    }
    if (!v?.token?.value && !v?.headers && !v?.codegenHeaders) {
      return [v.host];
    }
    const token = v?.token?.value && `${v?.token?.type} ${v?.token?.value}`.trim();
    const serverHeaders = typeof v?.headers?.serverOnly === "object" && v?.headers?.serverOnly;
    if (v?.headers?.serverOnly) {
      delete v.headers.serverOnly;
    }
    const headers = {
      ...v?.headers && { ...v.headers, ...serverHeaders },
      ...token && { [v.token.name]: token },
      ...v?.codegenHeaders
    };
    return [{ [v.host]: { headers } }];
  };
  const codegenConfig = {
    skipTypename: options?.skipTypename,
    useTypeImports: options?.useTypeImports,
    dedupeFragments: options?.dedupeFragments,
    gqlImport: "graphql-request#gql",
    onlyOperationTypes: options.onlyOperationTypes,
    namingConvention: {
      enumValues: "change-case-all#upperCaseFirst"
    }
  };
  const generates = Object.entries(options.clients).reduce((acc, [k, v]) => {
    const clientDocuments = options.documents.filter((file) => {
      const clientInExt = new RegExp(`\\.${k}\\.(gql|graphql)$`);
      const clientInPath = new RegExp(`\\/${k}\\/(?=${file.split("/").pop().replace(/\./g, "\\.")})`);
      return clientInExt.test(file) || clientInPath.test(file);
    });
    const noClientSpecified = options.documents.filter((file) => {
      const clientInExt = /\.\w+\.(gql|graphql)$/.test(file);
      const clientInPath = new RegExp(`\\/(${Object.keys(options.clients).join("|")})\\/(?=${file.split("/").pop().replace(/\./g, "\\.")})`).test(file);
      return !clientInExt && !clientInPath;
    });
    return {
      ...acc,
      [`${k}.ts`]: {
        config: codegenConfig,
        schema: prepareSchema(v),
        plugins: options.plugins,
        documents: k !== "default" ? clientDocuments : noClientSpecified
      }
    };
  }, {});
  return { silent: options.silent, generates };
}
async function generate(options) {
  const config = prepareConfig(options);
  return await generate$1(config, false);
}

const deepmerge = (a, b) => {
  const result = { ...a };
  for (const key in b) {
    if (typeof b[key] === "object" && b[key] !== null) {
      result[key] = deepmerge(result[key] || {}, b[key]);
    } else {
      result[key] = b[key];
    }
  }
  return result;
};

function prepareContext(ctx, prefix) {
  ctx.fns = Object.values(ctx.template).reduce((acc, template) => {
    const fns = template.match(/\w+\s*(?=\(variables)/g)?.sort() || [];
    return [...acc, ...fns];
  }, []);
  const fnName = (fn) => prefix + upperFirst(fn);
  const fnExp = (fn, typed = false) => {
    const name = fnName(fn);
    if (!typed) {
      const client = ctx?.clients.find((c) => ctx?.clientOps?.[c]?.includes(fn));
      return `export const ${name} = (...params) => GqlInstance().handle(${client ? `'${client}'` : ""})['${fn}'](...params)`;
    }
    return `  export const ${name}: (...params: Parameters<GqlSdkFuncs['${fn}']>) => ReturnType<GqlSdkFuncs['${fn}']>`;
  };
  ctx.generateImports = () => [
    "import { useGql } from '#imports'",
    ...ctx.clients.map((client) => `import { getSdk as ${client}GqlSdk } from '#gql/${client}'`),
    "export const GqlSdks = {",
    `    default: ${ctx.clients.find((c) => c === "default") || ctx.clients[0]}GqlSdk,`,
    ...ctx.clients.map((client) => `  ${client}: ${client}GqlSdk,`),
    "}",
    "const ctx = { instance: null }",
    "export const GqlInstance = () => {",
    " if (!ctx?.instance) {ctx.instance = useGql()}",
    " return ctx.instance",
    "}",
    `export const GqlOperations = ${JSON.stringify(ctx.clientOps)}`,
    ...ctx.fns.map((f) => fnExp(f))
  ].join("\n");
  ctx.generateDeclarations = () => [
    ...ctx.clients.map((client) => `import { getSdk as ${client}GqlSdk } from '#gql/${client}'`),
    ...Object.entries(ctx.clientTypes || {}).map(([k, v]) => genExport(`#gql/${k}`, v)),
    "declare module '#gql' {",
    `  type GqlClients = '${ctx.clients.join("' | '") || "default"}'`,
    "  const GqlOperations = {}",
    "  type GqlSdkValues<T extends GqlClients> = ReturnType<typeof GqlSdks[T]>",
    ...ctx.fns.map((f) => fnExp(f, true)),
    `  type GqlSdkFuncs = ${ctx.clients.map((c) => `ReturnType<typeof ${c}GqlSdk>`).join(" & ")}`,
    "}"
  ].join("\n");
  ctx.fnImports = ctx.fns.map((fn) => ({ from: "#gql", name: fnName(fn) }));
}
async function prepareOperations(ctx, path) {
  const scanFile = async (file) => {
    let clientToUse;
    const reExt = new RegExp(`\\.(${ctx.clients.join("|")})\\.(gql|graphql)$`);
    if (reExt.test(file)) {
      clientToUse = reExt.exec(file)?.[1];
    }
    const fileName = file.split("/").pop().replace(/\./g, "\\.");
    const reDir = new RegExp(`\\/(${ctx.clients.join("|")})\\/(?=${fileName})`);
    if (!clientToUse && reDir.test(file)) {
      clientToUse = reDir.exec(file)?.[1];
    }
    const { definitions } = parse(await promises.readFile(file, "utf8"));
    const operations = definitions.map(({ name }) => {
      if (!name?.value) {
        throw new Error(`Operation name missing in: ${file}`);
      }
      return name.value;
    });
    for (const op of operations) {
      clientToUse = new RegExp(`^(${ctx.clients.join("|")})(?=\\_)`).exec(op)?.[0] || clientToUse;
      if (!clientToUse || !ctx.clientOps?.[clientToUse]) {
        clientToUse = clientToUse || ctx.clients.find((c) => c === "default") || ctx.clients[0];
      }
      if (!ctx.clientOps?.[clientToUse]?.includes(op)) {
        ctx.clientOps[clientToUse].push(op);
      }
    }
  };
  for (const file of path) {
    await scanFile(file);
  }
}
function prepareTemplate(ctx) {
  ctx.clientTypes = ctx.clientTypes || {};
  ctx.clientTypes = Object.entries(ctx.template).reduce((acc, [key, template]) => {
    acc[key] = template.match(/^export\stype\s\w+(?=\s=\s)/gm).filter((e) => !["Scalars", "SdkFunctionWrapper", "Sdk"].some((f) => e.includes(f))).map((e) => e.replace("export type ", ""));
    return acc;
  }, {});
}

const logger = useLogger("nuxt-graphql-client");
const module = defineNuxtModule({
  meta: {
    name,
    version,
    configKey: "graphql-client",
    compatibility: {
      nuxt: "^3.0.0-rc.9"
    }
  },
  defaults: {
    clients: {},
    watch: true,
    autoImport: true,
    functionPrefix: "Gql"
  },
  async setup(opts, nuxt) {
    const resolver = createResolver(import.meta.url);
    const srcResolver = createResolver(nuxt.options.srcDir);
    nuxt.options.build.transpile.push(resolver.resolve("runtime"));
    const config = defu(
      {},
      nuxt.options.runtimeConfig.public["graphql-client"],
      nuxt.options.runtimeConfig.public.gql,
      opts
    );
    const codegenDefaults = {
      silent: true,
      skipTypename: true,
      useTypeImports: true,
      dedupeFragments: true,
      onlyOperationTypes: true
    };
    config.codegen = defu(config.codegen, codegenDefaults);
    const ctx = {
      clientOps: {},
      fnImports: [],
      clients: Object.keys(config.clients)
    };
    if (!ctx?.clients?.length) {
      const host = process.env.GQL_HOST || nuxt.options.runtimeConfig.public.GQL_HOST;
      const clientHost = process.env.GQL_CLIENT_HOST || nuxt.options.runtimeConfig.public.GQL_CLIENT_HOST;
      if (!host) {
        throw new Error("GQL_HOST is not set in public runtimeConfig");
      }
      ctx.clients = ["default"];
      config.clients = !clientHost ? { default: host } : { default: { host, clientHost } };
    }
    nuxt.options.runtimeConfig["graphql-client"] = { clients: {} };
    nuxt.options.runtimeConfig.public["graphql-client"] = defu(nuxt.options.runtimeConfig.public["graphql-client"], { clients: {} });
    const clientDefaults = {
      token: { type: "Bearer", name: "Authorization" },
      proxyCookies: true,
      preferGETQueries: config?.preferGETQueries ?? false
    };
    for (const [k, v] of Object.entries(config.clients)) {
      const conf = defu(typeof v !== "object" ? { host: v } : { ...v, token: typeof v.token === "string" ? { value: v.token } : v.token }, clientDefaults);
      const runtimeHost = k === "default" ? process.env.GQL_HOST : process.env?.[`GQL_${k.toUpperCase()}_HOST`];
      if (runtimeHost) {
        conf.host = runtimeHost;
      }
      const runtimeClientHost = k === "default" ? process.env.GQL_CLIENT_HOST : process.env?.[`GQL_${k.toUpperCase()}_CLIENT_HOST`];
      if (runtimeClientHost) {
        conf.clientHost = runtimeClientHost;
      }
      if (!conf?.host) {
        throw new Error(`GraphQL client (${k}) is missing it's host.`);
      }
      const runtimeToken = k === "default" ? process.env.GQL_TOKEN : process.env?.[`GQL_${k.toUpperCase()}_TOKEN`];
      if (runtimeToken) {
        conf.token.value = runtimeToken;
      }
      const runtimeTokenName = k === "default" ? process.env.GQL_TOKEN_NAME : process.env?.[`GQL_${k.toUpperCase()}_TOKEN_NAME`];
      if (runtimeTokenName) {
        conf.token.name = runtimeTokenName;
      }
      const schema = conf?.schema && srcResolver.resolve(conf.schema);
      if (schema && !existsSync(schema)) {
        delete conf.schema;
        logger.warn(`[nuxt-graphql-client] The Schema provided for the (${k}) GraphQL Client does not exist. \`host\` will be used as fallback.`);
      }
      ctx.clientOps[k] = [];
      config.clients[k] = deepmerge({}, conf);
      nuxt.options.runtimeConfig.public["graphql-client"].clients[k] = deepmerge({}, conf);
      if (conf?.token?.value) {
        nuxt.options.runtimeConfig["graphql-client"].clients[k] = { token: conf.token };
        if (!conf?.retainToken) {
          nuxt.options.runtimeConfig.public["graphql-client"].clients[k].token.value = void 0;
        }
      }
    }
    const documentPaths = [srcResolver.resolve()];
    if (config.documentPaths) {
      for (const path of config.documentPaths) {
        const dir = srcResolver.resolve(path);
        if (existsSync(dir)) {
          documentPaths.push(dir);
        } else {
          logger.warn(`[nuxt-graphql-client] Invalid document path: ${dir}`);
        }
      }
    }
    const gqlMatch = "**/*.{gql,graphql}";
    async function generateGqlTypes() {
      const documents = [];
      for await (const path of documentPaths) {
        const files = (await resolveFiles(path, [gqlMatch, "!**/schemas"], { followSymbolicLinks: false })).filter(allowDocument);
        documents.push(...files);
      }
      const plugins = ["typescript"];
      if (documents?.length) {
        plugins.push("typescript-operations", "typescript-graphql-request");
      }
      ctx.template = await generate({
        clients: config.clients,
        plugins,
        documents,
        resolver: srcResolver,
        ...typeof config.codegen !== "boolean" && config.codegen
      }).then((output) => output.reduce((acc, c) => ({ ...acc, [c.filename.split(".ts")[0]]: c.content }), {}));
      await prepareOperations(ctx, documents);
      prepareTemplate(ctx);
      prepareContext(ctx, config.functionPrefix);
    }
    addPlugin(resolver.resolve("runtime/plugin"));
    if (config.autoImport) {
      nuxt.options.alias["#gql"] = resolver.resolve(nuxt.options.buildDir, "gql");
      addTemplate({
        filename: "gql.mjs",
        getContents: () => ctx.generateImports()
      });
      addTemplate({
        filename: "gql/index.d.ts",
        getContents: () => ctx.generateDeclarations()
      });
      for (const client of ctx.clients) {
        addTemplate({
          write: true,
          filename: `gql/${client}.ts`,
          getContents: () => ctx.template[client]
        });
      }
      nuxt.hook("imports:extend", (autoimports) => {
        autoimports.push(...ctx.fnImports);
      });
      addImportsDir(resolver.resolve("runtime/composables"));
    }
    const allowDocument = (f) => {
      const isSchema = f.match(/([^/]+)\.(gql|graphql)$/)?.[0]?.toLowerCase().includes("schema");
      return !isSchema && !!statSync(srcResolver.resolve(f)).size;
    };
    if (config.watch) {
      nuxt.hook("builder:watch", async (event, path) => {
        if (!path.match(/\.(gql|graphql)$/)) {
          return;
        }
        if (event !== "unlink" && !allowDocument(path)) {
          return;
        }
        const start = Date.now();
        await generateGqlTypes();
        await nuxt.callHook("builder:generateApp");
        const time = Date.now() - start;
        logger.success(`[GraphQL Client]: Generation completed in ${time}ms`);
      });
    }
    await generateGqlTypes();
    extendViteConfig((config2) => {
      config2.optimizeDeps?.include?.push("graphql-request");
    });
  }
});

export { module as default };
